#include <queue>
#include <stack>
#include <fstream>
#include <iostream>

#include "BvhLoader.h"



namespace CHCDemo
{

using namespace std;


#define TYPE_INTERIOR -2
#define TYPE_LEAF -3



BvhLeaf *BvhLoader::ImportBinLeaf(ifstream &stream, BvhInterior *parent)
{
	BvhLeaf *leaf = new BvhLeaf(parent);

	int first, last;

	stream.read(reinterpret_cast<char *>(&first), sizeof(int));
	stream.read(reinterpret_cast<char *>(&last), sizeof(int));

	leaf->mFirst = first;
	leaf->mLast = last;

	return leaf;
}


BvhInterior *BvhLoader::ImportBinInterior(ifstream &stream, BvhInterior *parent)
{
	BvhInterior *interior = new BvhInterior(parent);

	stream.read(reinterpret_cast<char *>(&interior->mAxis), sizeof(char));
	stream.read(reinterpret_cast<char *>(&interior->mPosition), sizeof(float));

	return interior;
}


BvhNode *BvhLoader::LoadNextNode(ifstream &stream, BvhInterior *parent)
{
	int nodeType;
	stream.read(reinterpret_cast<char *>(&nodeType), sizeof(int));

	if (nodeType == TYPE_LEAF)
	{
		//OUT1("l");
		return ImportBinLeaf(stream, static_cast<BvhInterior *>(parent));
	}

	if (nodeType == TYPE_INTERIOR)
	{
		//OUT1("i");
		return ImportBinInterior(stream, static_cast<BvhInterior *>(parent));
	}

	return NULL;
}


Bvh *BvhLoader::LoadFromFile(const string &filename,  
							 const SceneEntityContainer &entities)
{
	// export binary version of mesh
	queue<BvhNode *> tStack;
	ifstream stream(filename.c_str(), ifstream::binary);

	if (!stream.is_open()) return NULL;

	cout << "loading bvh" << endl;
	Bvh *bvh = new Bvh();

	bvh->mGeometrySize = entities.size();
	bvh->mGeometry = new SceneEntity*[bvh->mGeometrySize];

	for (size_t i = 0; i < bvh->mGeometrySize; ++ i)
		bvh->mGeometry[i] = entities[i];

	bvh->mRoot = LoadNextNode(stream, NULL);

	tStack.push(bvh->mRoot);
	bvh->mNumNodes = 1;

	while(!tStack.empty())
	{
		BvhNode *node = tStack.front();
		tStack.pop();

		if (!node->IsLeaf())
		{
			bvh->mNumNodes += 2;

			BvhInterior *interior = static_cast<BvhInterior *>(node);

            BvhNode *front = LoadNextNode(stream, interior);
			BvhNode *back = LoadNextNode(stream, interior);
	
			interior->mFront = front;
			interior->mBack = back;

			front->mDepth = interior->mDepth + 1;
			back->mDepth = interior->mDepth + 1;

			tStack.push(front);			
			tStack.push(back);
		}
	}

	cout << "... finished loading " << bvh->mNumNodes << " nodes, updating boxes" << endl;

	//adjust bounding boxes
	//bvh->UpdateBoxes(bvh->mRoot);
	//bvh->UpdateNumLeaves(bvh->mRoot);

	// compute unique ids
	//bvh->ComputeIds();
	// update the indices of the geometry so we can render interiors as well
	//bvh->UpdateInteriors(bvh->mRoot);
	// do this once so at least the current node bounding box is tested
	//bvh->RecomputeBounds();

	return bvh;
}


}
